module Generator
  class GeneratorError < StandardError; end

  class Token
    def initialize(type, src = nil, dst = nil)
      @type, @src, @dst = type, src, dst
    end

    def need_imm?
    end

    def to_code
      case @type
      when :Print
        [ "movq #{@src}, %rsi", "movq $__print_fmt, %rdi", "movq $0, %rax", "call printf" ]
      when :Move
        [ @src.is_a?(Integer) ? "movq $#{@src}, %r11" : "movq #{@src}, %r11", "movq %r11, #{@dst}" ]
      when :Tilde
        [ "movq #{@src}, %r11", "notq %r11", "movq %r11, #{@dst}" ]
      when :Negative
        [ "movq #{@src}, %r11", "negq %r11", "movq %r11, #{@dst}" ]
      when :Plus
        [ "movq #{@src[0]}, %r11", "movq %r11, #{@dst}", "movq #{@src[1]}, %r11", "addq %r11, #{@dst}"]
      when :Minus
        [ "movq #{@src[0]}, %r11", "movq %r11, #{@dst}", "movq #{@src[1]}, %r11", "subq %r11, #{@dst}"]
      when :Asterisk
        [ "movq #{@src[0]}, %rax", "movq #{@src[1]}, %rcx", "imulq %rcx", "movq %rax, #{@dst}" ]
      when :Slash
        [ "movq #{@src[0]}, %rax", "movq #{@src[1]}, %rcx", "cqo", "idivq %rcx", "movq %rax, #{@dst}" ]
      when :Percent
        [ "movq #{@src[0]}, %rax", "movq #{@src[1]}, %rcx", "cqo", "idivq %rcx", "movq %rdx, #{@dst}" ]
      when :Bar
        [ "movq #{@src[0]}, %rax", "movq #{@src[1]}, %rcx", "orq %rcx, %rax", "movq %rax, #{@dst}" ]
      when :Ampersand
        [ "movq #{@src[0]}, %rax", "movq #{@src[1]}, %rcx", "andq %rcx, %rax", "movq %rax, #{@dst}" ]
      when :Caret
        [ "movq #{@src[0]}, %rax", "movq #{@src[1]}, %rcx", "xorq %rcx, %rax", "movq %rax, #{@dst}" ]
      when :LeftShift
        [ "movq #{@src[0]}, %rax", "movq #{@src[1]}, %rcx", "salq %cl, %rax", "movq %rax, #{@dst}" ]
      when :RightShift
        [ "movq #{@src[0]}, %rax", "movq #{@src[1]}, %rcx", "sarq %cl, %rax", "movq %rax, #{@dst}" ]
      end
    end
  end

  BinaryOps = {
    :Plus         => 50,
    :Minus        => 50,
    :Asterisk     => 60,
    :Slash        => 60,
    :Percent      => 60,
    :Bar          => 10,
    :Ampersand    => 30,
    :Caret        => 20,
    :LeftShift    => 40,
    :RightShift   => 40,
  }

  UnaryOps = {
    :Tilde        => 70,
    :Negative     => 80,
  }

  Vars = Array.new

  def self.var_temp
    #var = "#{8 * Vars.length}(%rsp)"
    var = "tmp_#{Vars.length}"
    Vars << var unless Vars.include?(var)
    var
  end

  def self.generate_expression(token)
    if token.type == :Number
      t = var_temp
      [t, [Token.new(:Move, token.child[0], t)]]
    elsif token.type == :Variable
      [token.child[0], []]
    elsif token.type == :Bracket
      var, array = generate_expression(token.child[0])
      t = var_temp
      [t, [array, Token.new(:Move, var, t)]]
    elsif UnaryOps.include?(token.type)
      var, array = generate_expression(token.child[0])
      t = var_temp
      [t, [array, Token.new(token.type, var, t)]]
    else
      var1, array1 = generate_expression(token.child[0])
      var2, array2 = generate_expression(token.child[1])
      t = var_temp
      [t, [array1, array2, Token.new(token.type, [var1, var2], t)]]
    end
  end

  def self.generate_statement(token)
    array = Array.new
    if token.type == :Print
      var, garray = generate_expression(token.child[0])
      array << garray unless garray.empty?
      array << Token.new(:Print, var)
    elsif token.type == :Move
      Vars << token.child[0].child[0] unless Vars.include?(token.child[0].child[0])
      var, garray = generate_expression(token.child[1])
      array << garray unless garray.empty?
      array << Token.new(:Move, var, token.child[0].child[0])
    end
    array
  end

  def self.generate(token)
    array = Array.new
    token.child.each do |statement|
      array << generate_statement(statement)
    end

    output = %{
  .section .data
__print_fmt:
  .string "%ld\\n"

  .section .text
  .globl main
  main:
  pushq %rbp
  movq %rsp, %rbp
  subq <VARIABLE>, %rsp

<CODE>
  movq %rbp, %rsp
  popq %rbp
  movq $0, %rax
  retq
}
    text = ""
    array.flatten.each { |ops| ops.to_code.each { |line| text += " " * 2 + line + "\n" } }

    Vars.length.times do |i|
      regex = "(?<![\\w])#{Vars[i]}(?![\\w])"
      text.gsub!(/#{regex}/, i > 0 ? "#{8*i}(%rsp)" : "(%rsp)")
    end
    output.gsub('<VARIABLE>', "$#{Vars.length * 8}").gsub('<CODE>', text)
  end
end
